import os
import numpy as np
import tensorflow.compat.v2 as tf
from tf_agents.replay_buffers import tf_uniform_replay_buffer
from tf_agents.specs import tensor_spec
from tf_agents.utils import common
import gym
import random
from tqdm import tqdm
from absl import flags
from absl import app

from algae_dice import algae
from algae_dice import wrappers


import envs

seed = 42
tf.enable_v2_behavior()
tf.random.set_seed(seed)
np.random.seed(seed)
random.seed(seed)
max_episode_steps = int(3000)

base_dir = './copo_bipedal_results'

eval_env = gym.make("ConstrainedBipedal-v0")
eval_env = wrappers.check_and_normalize_box_actions(eval_env)
eval_env.seed(seed + 15)

# init_spec = tensor_spec.TensorSpec([env.observation_space.shape[0]],
#                                 tf.float32, 'observation')


# Set up replay buffer iterators
# replay_buffer_iter = iter(
#     replay_buffer.as_dataset(sample_batch_size=FLAGS.sample_batch_size))

# init_replay_buffer_iter = iter(
#     init_replay_buffer.as_dataset(sample_batch_size=FLAGS.sample_batch_size))

def eval(rl_algo):
    # Evaluate the trained policy
    average_returns, evaluation_timesteps, average_per_step, average_costs, average_ps_cost = rl_algo.evaluate(
            eval_env, 0.99, max_episode_steps=max_episode_steps)
    return [average_returns, evaluation_timesteps, average_per_step, average_costs, average_ps_cost]

def collect_data(exp_idx, env):
    # os.makedirs(os.path.join(base_dir,'experiment'+str(exp_idx)))
    spec = (
        tensor_spec.TensorSpec([env.observation_space.shape[0]], tf.float32,
                                'observation'),
        tensor_spec.TensorSpec([env.action_space.shape[0]], tf.float32, 'action'),
        tensor_spec.TensorSpec([env.observation_space.shape[0]], tf.float32,
                                'next_observation'),
        tensor_spec.TensorSpec([1], tf.float32, 'reward'),
        tensor_spec.TensorSpec([1], tf.float32, 'cost'),
        tensor_spec.TensorSpec([1], tf.float32, 'mask'),
        tensor_spec.TensorSpec([1], tf.float32, 'timestep'),
    )
    max_projection_steps = 1000000
    replay_buffer = tf_uniform_replay_buffer.TFUniformReplayBuffer(
        spec, batch_size=1, max_length=max_projection_steps)
    rb_checkpointer = common.Checkpointer(
        ckpt_dir=os.path.join(base_dir,'experiment' + str(exp_idx), 'replay_buffer_proj'),
        max_to_keep=1,
        replay_buffer=replay_buffer)
    rb_checkpointer.initialize_or_restore()
    total_timesteps, episode_return, episode_timesteps, done = 0, 0, 0, True
    rl_algo = algae.ALGAE(
        env.observation_space.shape[0],
        env.action_space.shape[0],
        int(1000),
        critic_lr=5e-6,
        actor_lr=5e-6,
        use_dqn=True,
        use_init_states=True,
        algae_alpha=0.01,
        exponent=1.5)
    rl_algo.load_model('./algae_bipedal_chkp/algae_model')
    while total_timesteps < max_projection_steps:
        if done:
            obs, episode_return, episode_timesteps = env.reset(), 0, 0
        _, action, _ = rl_algo.actor(np.array([obs]))
        action = action[0].numpy()
        next_obs, reward, done, info = env.step(action)
        if np.any(np.isnan(next_obs)):
            print("NaNs in states while collecting")
            exit()
        cost = info.get('cost', 0)
        if (max_episode_steps is not None and
            episode_timesteps + 1 == max_episode_steps):
            done = True
        if not done or episode_timesteps + 1 == max_episode_steps:  # pylint: disable=protected-access
            mask = 1.0
        else:
            mask = 0.0
        replay_buffer.add_batch((np.array([obs.astype(np.float32)]),
                                np.array([action.astype(np.float32)]),
                                np.array([next_obs.astype(np.float32)]),
                                np.array([[reward]]).astype(np.float32),
                                np.array([[cost]]).astype(np.float32),
                                np.array([[mask]]).astype(np.float32),
                                np.array([[episode_timesteps]]).astype(np.float32)))
        episode_return += reward
        episode_timesteps += 1
        total_timesteps += 1
        obs = next_obs
    rb_checkpointer.save(global_step=total_timesteps)

def train_agents(exp_idx):
    results_list = []
    env = gym.make("ConstrainedBipedal-v0")
    env = wrappers.check_and_normalize_box_actions(env)
    rl_algo = algae.ALGAE(
        env.observation_space.shape[0],
        env.action_space.shape[0],
        int(1000),
        critic_lr=5e-6,
        actor_lr=5e-6,
        use_dqn=True,
        use_init_states=True,
        algae_alpha=0.01,
        exponent=1.5)
    rl_algo.load_model('./algae_bipedal_chkp/algae_model')

    spec = (
        tensor_spec.TensorSpec([env.observation_space.shape[0]], tf.float32,
                                'observation'),
        tensor_spec.TensorSpec([env.action_space.shape[0]], tf.float32, 'action'),
        tensor_spec.TensorSpec([env.observation_space.shape[0]], tf.float32,
                                'next_observation'),
        tensor_spec.TensorSpec([1], tf.float32, 'reward'),
        tensor_spec.TensorSpec([1], tf.float32, 'cost'),
        tensor_spec.TensorSpec([1], tf.float32, 'mask'),
        tensor_spec.TensorSpec([1], tf.float32, 'timestep'),
    )
    max_projection_steps = 1000000
    replay_buffer = tf_uniform_replay_buffer.TFUniformReplayBuffer(
        spec, batch_size=1, max_length=max_projection_steps)
    replay_buffer_iter = iter(
        replay_buffer.as_dataset(sample_batch_size=256))
    rb_checkpointer = common.Checkpointer(
        ckpt_dir=os.path.join(base_dir,'experiment' + str(exp_idx), 'replay_buffer_proj'),
        max_to_keep=1,
        replay_buffer=replay_buffer)
    rb_checkpointer.initialize_or_restore()
    model_dir = os.path.join(base_dir,'experiment' + str(exp_idx),"copo_bipedal_projection_chkp")
    max_to_keep = 10
    rl_algo.setup_checkpointers(model_dir, max_to_keep)
    results_list.append(eval(rl_algo))
    def transport(s,a,ns,na):
        return tf.norm(s[:,:14] - ns[:,:14], ord=2) + tf.norm(a-na, ord=2)
    rl_algo.init_copo(20.0, 500.0, transport, 0.4)
    target_entropy = (-env.action_space.shape[0])
    for train_step in range(8000): 
        try:
            rl_algo.trainCOPO(
                replay_buffer_iter,
                discount=1.0,
                tau=0.005,
                target_entropy=target_entropy,
                actor_update_freq=5)
            if train_step % (10000 // 10) == 0:
                rl_algo.save_checkpointers()
                results_list.append(eval(rl_algo))
        except tf.errors.InvalidArgumentError as e:
            # print(f"User entered ctrl-c ... stopping training")
            # print(f"stopping training after {train_step} steps")
            return None
    results_list.append(eval(rl_algo))
    np.savetxt(os.path.join(base_dir,'experiment' + str(exp_idx), 'copo_results_' + str(exp_idx) + '.csv'),
                np.array(results_list), delimiter=',')

def create_env(i):
    env = gym.make("ConstrainedBipedal-v0")
    env = wrappers.check_and_normalize_box_actions(env)
    env.seed(i)
    return env

if __name__ == '__main__':
    num_experiments = 10
    
    envs = [create_env(i) for i in range(num_experiments)]
    with tqdm(total=num_experiments,desc="Collecting data ...") as pbar:
        for (i,e) in enumerate(envs):
            collect_data(i, e)
            pbar.update(1)

    with tqdm(total=num_experiments,desc="Training agents ...") as pbar:
        for i in range(0,num_experiments):
            train_agents(i)
            pbar.update(1)
